@@ -4,7 +4,7 @@ use warnings;
 use POSIX qw(:errno_h);
 use Socket qw(PF_INET SOCK_STREAM pack_sockaddr_in inet_ntoa $CRLF SOL_SOCKET SO_ERROR);
 use Fcntl qw(F_GETFL F_SETFL O_NONBLOCK);
-our $VERSION = "0.13";
+our $VERSION = "0.14";
 
 sub Hijk::Error::CONNECT_TIMEOUT () { 1 << 0 } # 1
 sub Hijk::Error::READ_TIMEOUT    () { 1 << 1 } # 2
@@ -22,14 +22,16 @@ sub read_http_message {
     $read_timeout = selectable_timeout($read_timeout);
     my ($body,$buf,$decapitated,$nbytes,$proto);
     my $status_code = 0;
+    my $no_content_len = 0;
     vec(my $rin = '', $fd, 1) = 1;
     do {
         my $nfound = select($rin, undef, undef, $read_timeout);
-
         return (undef,0,undef,undef, Hijk::Error::READ_TIMEOUT)
             if ($nfound != 1 || (defined($read_timeout) && $read_timeout <= 0));
 
         my $nbytes = POSIX::read($fd, $buf, $block_size);
+        return ($proto, $status_code, $body, $header)
+            if $no_content_len && $decapitated && (!defined($nbytes) || $nbytes == 0);
         if (!defined($nbytes)) {
             next
                 if ($! == EWOULDBLOCK || $! == EAGAIN);
@@ -41,7 +43,9 @@ sub read_http_message {
 
         if ($decapitated) {
             $body .= $buf;
-            $block_size -= $nbytes;
+            if (!$no_content_len) {
+                $block_size -= $nbytes;
+            }
         }
         else {
             $head .= $buf;
@@ -58,21 +62,99 @@ sub read_http_message {
                     my ($key, $value) = split /: /, $_, 2;
                     $header->{$key} = $value;
                 }
+                if ($header->{'Transfer-Encoding'} && $header->{'Transfer-Encoding'} eq 'chunked') {
+                    # if there is chunked encoding we have to ignore content lenght even if we have it
+                    return ($proto, $status_code, read_chunked_body($body, $fd, $read_timeout,$header), $header);
+                }
 
                 if ($header->{'Content-Length'}) {
                     $block_size = $header->{'Content-Length'} - length($body);
                 }
                 else {
-                    $block_size = 0;
+                    $block_size = 10204;
+                    $no_content_len = 1;
                 }
             }
         }
 
-    } while( !$decapitated || $block_size > 0 );
-
+    } while( !$decapitated || $block_size > 0 || $no_content_len);
     return ($proto, $status_code, $body, $header);
 }
 
+sub read_chunked_body {
+    my ($buf,$fd, $read_timeout,$header) = @_;
+    my $chunk_size   = 0;
+    my $body         = "";
+    my $block_size = 10240;
+    my $trailer_mode  = 0;
+
+    vec(my $rin = '', $fd, 1) = 1;
+    while(1) {
+        # just read a 10k block and process it until it is consumed
+        if (length($buf) == 0 || length($buf) < $chunk_size) {
+            my $nfound = select($rin, undef, undef, $read_timeout);
+            return (undef,0,undef,undef, Hijk::Error::READ_TIMEOUT)
+                if ($nfound != 1 || (defined($read_timeout) && $read_timeout <= 0));
+            my $current_buf = "";
+            my $nbytes = POSIX::read($fd, $current_buf, $block_size);
+            if (!defined($nbytes)) {
+                next
+                    if ($! == EWOULDBLOCK || $! == EAGAIN);
+                die "Failed to read chunked body from socket. errno = $!"
+            }
+
+            die "Failed to read chunked body from socket. Got 0 bytes back, which shouldn't happen <$buf> <$current_buf>"
+                if $nbytes == 0;
+            $buf .= $current_buf;
+        }
+        if ($trailer_mode) {
+            # http://tools.ietf.org/html/rfc2616#section-14.40
+            # http://tools.ietf.org/html/rfc2616#section-3.6.1
+            #   A server using chunked transfer-coding in a response MUST NOT use the
+            #   trailer for any header fields unless at least one of the following is
+            #   true:
+
+            #   a)the request included a TE header field that indicates "trailers" is
+            #     acceptable in the transfer-coding of the  response, as described in
+            #     section 14.39; or,
+
+            #   b)the server is the origin server for the response, the trailer
+            #     fields consist entirely of optional metadata, and the recipient
+            #     could use the message (in a manner acceptable to the origin server)
+            #     without receiving this metadata.  In other words, the origin server
+            #     is willing to accept the possibility that the trailer fields might
+            #     be silently discarded along the path to the client.
+
+            # in case of trailer mode, we just read everything until the next CRLFCRLF
+            my $neck_pos = index($buf, "${CRLF}${CRLF}");
+            if ($neck_pos > 0) {
+                return $body;
+            }
+        } else {
+            if ($chunk_size > 0 && length($buf) >= $chunk_size) {
+                $body .= substr($buf, 0, $chunk_size - 2); # our chunk size includes the CRLF
+                $buf = substr($buf, $chunk_size);
+                $chunk_size = 0;
+            } else {
+                my $neck_pos = index($buf, ${CRLF});
+                if ($neck_pos > 0) {
+                    $chunk_size = hex(substr($buf, 0, $neck_pos));
+                    if ($chunk_size == 0) {
+                        if ($header->{Trailer}) {
+                            $trailer_mode = 1;
+                        } else {
+                            return $body;
+                        }
+                    } else {
+                        $chunk_size += 2;                  # include the final CRLF
+                        $buf = substr($buf, $neck_pos + 2);
+                    }
+                }
+            }
+        }
+    }
+}
+
 sub construct_socket {
     my ($host, $port, $connect_timeout) = @_;
 
